#/*
 * auditor.c
 * Copyright (C) 2005 Robin KAY. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 */

#include "compat.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/mman.h>
#include <link.h>
#include <unistd.h>

#include "utils.h"

static const char *rules_env;

static void *base = NULL;
static off_t offset;

uint_t la_version(uint_t version)
{
  // Read the FSREWRITE_RULES environment variable
  rules_env = getenv("FSREWRITE_RULES");
  if (!rules_env) {
    fprintf(stderr, "fsrewrite: FSREWRITE_RULES is not set.\n");
    abort();
  }

  return LAV_CURRENT;
}

static void *grab_memory(size_t size)
{
  int pagesize = getpagesize();
  off_t old_offset;

  if (size > pagesize) {
    return NULL;
  }

  if (!base || offset + size > pagesize) {
    // mmap(2) a new page of anonymous memory
    base = mmap(
      NULL, pagesize, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON, -1, 0);
    if (base == MAP_FAILED) {
      fprintf(stderr, "fsrewrite: mmap(2) failed.");
      abort();
    }
    offset = 0;
  }

  old_offset = offset;
  offset += size;

  return (char *)base + old_offset;
}

typedef struct {
  const char *nname;
} objsearch_params_t;

static int objsearch_cb(const char *path, const char *vpath, void *cb_data)
{
  objsearch_params_t *params = (objsearch_params_t *)cb_data;
  int retval;

  retval = access(path, F_OK);
  if (retval >= 0) {
    if (strcmp(path, vpath) == 0) {
      // Pass through vpaths which are unchanged by path_rewrite
      params->nname = vpath;
    }
    else {
      char *nname;
      int path_len;

      // The life-cycle of values returned by la_objsearch isn't clearly 
      // specified, so the memory allocated here is never freed. 
      // grab_memory() cuts grabs sections from mmap(2)ed chunks of
      // anonymous memory since malloc(3C) may not be available.
      path_len = strlen(path) + 1;
      nname = grab_memory(path_len);
      memcpy(nname, path, path_len);
      params->nname = nname;
    }
  }
  return retval;
}

char *la_objsearch(const char *name, uintptr_t *cookie, uint_t flags)
{
  if (name[0] == '/') {
    objsearch_params_t params;

    params.nname = NULL;
    path_rewrite(name, rules_env, &objsearch_cb, &params);
    return (char *)params.nname;
  }
  else {
    // Pass through names which are not absolute paths
    return (char *)name;
  }
}
